using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Reflection;

namespace SimpleDoubleDispatchSample {
    public class Program {
        private static void Main(string[] args) {
            Collection<AbstractVisitor> visitors = new Collection<AbstractVisitor>();
            Collection<Pet> pets = new Collection<Pet>();

            pets.Add(new Fish());
            pets.Add(new Dog());

            visitors.Add(new Feeder());
            visitors.Add(new Walker());

            // Visit each of the pets.
            foreach (Pet pet in pets) {
                foreach (AbstractVisitor visitor in visitors) {
                    visitor.Visit(pet);
                }
            }

            // Check the results.
            foreach(Pet pet in pets) {
                Console.Out.WriteLine(pet.GetType().Name);
                foreach(String note in pet.Visitors) {
                    Console.Out.WriteLine("\t{0}", note);
                }
            }
        }
    }

    /// <summary>
    /// Concrete visitor, a pet feeder.
    /// </summary>
    public class Feeder : AbstractVisitor {
        public void Visit(Dog dog) {
            // Feed the dog
            dog.Visitors.Add("Fed the dog");
        }

        public void Visit(Fish fish) {
            // Feed the fish
            fish.Visitors.Add("Fed the fish");
        }
    }

    /// <summary>
    /// Concrete visitor, a pet walker.
    /// </summary>
    public class Walker : AbstractVisitor {
        public void Visit(Dog dog) {
            dog.Visitors.Add("Walked the dog");
        }

        // Fish can't be walked!
    }

    /// <summary>
    /// Base class for pets.
    /// </summary>
    public abstract class Pet {
        private readonly Collection<string> visitors;

        public Collection<string> Visitors {
            get { return visitors; }
        }

        protected Pet() {
            visitors = new Collection<string>();
        }
    }

    /// <summary>
    /// A pet fish.
    /// </summary>
    public class Fish : Pet {}

    /// <summary>
    /// A pet doc.
    /// </summary>
    public class Dog : Pet {}

    /// <summary>
    /// Abstract base class for visitors.
    /// </summary>
    public abstract class AbstractVisitor {
        private static readonly Dictionary<long, MethodInfo> dispatch;

        static AbstractVisitor() {
            dispatch = new Dictionary<long, MethodInfo>();

            foreach (Type t in Assembly.GetCallingAssembly().GetTypes()) {
                if (t.IsSubclassOf(typeof(AbstractVisitor))) {
                    foreach (MethodInfo mi in t.GetMethods()) {
                        if (mi.Name == "Visit") {
                            ParameterInfo[] pars = mi.GetParameters();
                            if (pars.Length == 1) {
                                Int64 code = ((Int64)t.GetHashCode() << 32) +
                                    pars[0].ParameterType.GetHashCode();
                                dispatch.Add(code, mi);
                            }
                        }
                    }
                }
            }
        }

        public virtual void Visit(object pet) {
            Int64 hash = ((Int64)GetType().GetHashCode() << 32) + pet.GetType().GetHashCode();
            if (dispatch.ContainsKey(hash)) {
                dispatch[hash].Invoke(this, new object[] { pet });
            }
            else {
                // This is our fallback functionality
                Console.WriteLine("Visiting object " + pet.ToString());
            }
        }
    }
}